@@ -1,21 +1,21 @@
 /* defs.h -- data types and declarations.
-   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2001, 2003, 2004, 2005, 2006, 2007 Free Software Foundation, Inc.
+   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2004 Free Software Foundation, Inc.
 
-   This program is free software: you can redistribute it and/or modify
+   This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation, either version 3 of the License, or
-   (at your option) any later version.
-   
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+   USA.
 */
-
-
 #ifndef INC_DEFS_H
 #define INC_DEFS_H 1
 
@@ -83,9 +83,14 @@ int stat PARAMS((const char *__path, struct stat *__statbuf));
 int optionl_stat PARAMS((const char *name, struct stat *p));
 int optionp_stat PARAMS((const char *name, struct stat *p));
 int optionh_stat PARAMS((const char *name, struct stat *p));
+int debug_stat   PARAMS((const char *file, struct stat *bufp));
 
 int get_statinfo PARAMS((const char *pathname, const char *name, struct stat *p));
 
+#if ! defined HAVE_FCHDIR && ! defined fchdir
+# define fchdir(fd) (-1)
+#endif
+
 
 
 #ifndef S_ISUID
@@ -195,7 +200,7 @@ struct long_val
 struct perm_val
 {
   enum permissions_type kind;
-  mode_t val[2];
+  mode_t val;
 };
 
 /* dir_id is used to support loop detection in find.c and 
@@ -303,6 +308,9 @@ struct predicate
   /* True if this predicate node requires knowledge of the file type. */
   boolean need_type;
 
+  /* True if this predicate should display control characters literally */
+  boolean literal_control_chars;
+  
   /* Information needed by the predicate processor.
      Next to each member are listed the predicates that use it. */
   union
@@ -337,9 +345,9 @@ struct predicate
   const struct parser_table* parser_entry;
 };
 
-/* find.c. */
-int get_info PARAMS((const char *pathname, const char *name, struct stat *p, struct predicate *pred_ptr));
-int following_links(void);
+/* find.c, ftsfind.c */
+boolean is_fts_enabled();
+
 
 
 /* find library function declarations.  */
@@ -351,7 +359,6 @@ char *dirname PARAMS((char *path));
 void error PARAMS((int status, int errnum, char *message, ...));
 
 /* listfile.c */
-void list_file PARAMS((char *name, char *relname, struct stat *statp, time_t current_time, int output_block_size, FILE *stream));
 char *get_link_name PARAMS((char *name, char *relname));
 
 /* stpcpy.c */
@@ -431,7 +438,7 @@ boolean pred_amin PARAMS((char *pathname, struct stat *stat_buf, struct predicat
 boolean pred_and PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_anewer PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_atime PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_closeparen PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
+boolean pred_close PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_cmin PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_cnewer PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_comma PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
@@ -440,6 +447,7 @@ boolean pred_delete PARAMS((char *pathname, struct stat *stat_buf, struct predic
 boolean pred_empty PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_exec PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_execdir PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
+boolean pred_executable PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_false PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_fls PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_fprint PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
@@ -464,7 +472,7 @@ boolean pred_nogroup PARAMS((char *pathname, struct stat *stat_buf, struct predi
 boolean pred_nouser PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_ok PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_okdir PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
-boolean pred_openparen PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
+boolean pred_open PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_or PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_path PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_perm PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
@@ -472,6 +480,7 @@ boolean pred_print PARAMS((char *pathname, struct stat *stat_buf, struct predica
 boolean pred_print0 PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_prune PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_quit PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
+boolean pred_readable PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_regex PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_samefile PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_size PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
@@ -480,6 +489,7 @@ boolean pred_type PARAMS((char *pathname, struct stat *stat_buf, struct predicat
 boolean pred_uid PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_used PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_user PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
+boolean pred_writable PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_xtype PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 
 
@@ -511,10 +521,15 @@ struct predicate *get_new_pred_chk_op PARAMS((const struct parser_table *entry))
 struct predicate *insert_primary PARAMS((const struct parser_table *entry));
 struct predicate *insert_primary_withpred PARAMS((const struct parser_table *entry, PRED_FUNC fptr));
 void usage PARAMS((char *msg));
+extern boolean check_nofollow(void);
+extern void complete_pending_execs(struct predicate *p);
+extern void complete_pending_execdirs(struct predicate *p);
+/* find.c. */
+int get_info PARAMS((const char *pathname, const char *name, struct stat *p, struct predicate *pred_ptr));
+int following_links(void);
+int digest_mode(mode_t mode, const char *pathname, const char *name, struct stat *pstat, boolean leaf);
+boolean default_prints (struct predicate *pred);
 
-extern char *program_name;
-extern struct predicate *predicates;
-extern struct predicate *last_pred;
 
 struct options
 {
@@ -538,9 +553,14 @@ struct options
    * no longer exists by the time we get around to processing it.
    */
   boolean ignore_readdir_race;
+
+  /* If true, pass control characters through.  If false, escape them
+   * or turn them into harmless things.
+   */
+  boolean literal_control_chars;
   
-/* If true, we issue warning messages
- */
+  /* If true, we issue warning messages
+   */
   boolean warnings;
   time_t start_time;		/* Time at start of execution.  */
   
@@ -590,8 +610,8 @@ struct state
      Used for stat, readlink, remove, and opendir.  */
   char *rel_pathname;
 
-  /* Length of current path. */
-  int path_length;
+  /* Length of starting path. */
+  int starting_path_length;
 
   /* If true, don't descend past current directory.
      Can be set by -prune, -maxdepth, and -xdev/-mount. */
@@ -600,12 +620,15 @@ struct state
   /* Status value to return to system. */
   int exit_status;
 };
-extern struct state state;
 
+/* finddata.c */
+extern struct state state;
 extern char const *starting_dir;
 extern int starting_desc;
-#if ! defined HAVE_FCHDIR && ! defined fchdir
-# define fchdir(fd) (-1)
-#endif
+extern struct predicate *eval_tree;
+extern char *program_name;
+extern struct predicate *predicates;
+extern struct predicate *last_pred;
+
 
 #endif
