@@ -1,11 +1,11 @@
 /* parser.c -- convert the command line args into an expression tree.
    Copyright (C) 1990, 1991, 1992, 1993, 1994, 2000, 2001, 2003, 
-                 2004, 2005, 2007 Free Software Foundation, Inc.
+                 2004, 2005 Free Software Foundation, Inc.
 
-   This program is free software: you can redistribute it and/or modify
+   This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation, either version 3 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -13,7 +13,9 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+   USA.
 */
 
 
@@ -32,7 +34,6 @@
 #include "nextelem.h"
 #include "stdio-safer.h"
 #include "regextype.h"
-#include "gnulib-version.h"
 
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
@@ -91,6 +92,7 @@ static boolean parse_depth         PARAMS((const struct parser_table*, char *arg
 static boolean parse_empty         PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_exec          PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_execdir       PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
+static boolean parse_executable    PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_false         PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_fls           PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_fprintf       PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
@@ -129,9 +131,11 @@ static boolean parse_perm          PARAMS((const struct parser_table*, char *arg
 static boolean parse_print0        PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_printf        PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_prune         PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
+static boolean parse_readable      PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_regex         PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_regextype     PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_samefile      PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
+static boolean parse_show_control_chars PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_size          PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_true          PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_type          PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
@@ -144,15 +148,13 @@ static boolean parse_xdev          PARAMS((const struct parser_table*, char *arg
 static boolean parse_ignore_race   PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_noignore_race PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_warn          PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
+static boolean parse_writable      PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_xtype         PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 static boolean parse_quit          PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 
-
-
 boolean parse_print             PARAMS((const struct parser_table*, char *argv[], int *arg_ptr));
 
 
-
 static boolean insert_type PARAMS((char **argv, int *arg_ptr, const struct parser_table *entry, PRED_FUNC which_pred));
 static boolean insert_regex PARAMS((char *argv[], int *arg_ptr, const struct parser_table *entry, int regex_options));
 static boolean insert_fprintf PARAMS((FILE *fp, const struct parser_table *entry, PRED_FUNC func, char *argv[], int *arg_ptr));
@@ -203,8 +205,8 @@ static struct parser_table const parse_table[] =
 {
  PARSE_PUNCTUATION("!",                     negate),
   PARSE_PUNCTUATION("not",                   negate),	     /* GNU */
-  PARSE_PUNCTUATION("(",                     openparen),
-  PARSE_PUNCTUATION(")",                     closeparen),
+  PARSE_PUNCTUATION("(",                     open),
+  PARSE_PUNCTUATION(")",                     close),
   PARSE_PUNCTUATION(",",                     comma),	     /* GNU */
   PARSE_PUNCTUATION("a",                     and),
   PARSE_TEST       ("amin",                  amin),	     /* GNU */
@@ -220,6 +222,7 @@ static struct parser_table const parse_table[] =
   PARSE_OPTION     ("depth",                 depth),
   PARSE_TEST       ("empty",                 empty),	     /* GNU */
   {ARG_ACTION,      "exec",    parse_exec, pred_exec}, /* POSIX */
+  PARSE_TEST       ("executable",            executable),    /* GNU, 4.3.0+ */
   PARSE_ACTION     ("execdir",               execdir), /* *BSD, GNU */
   PARSE_ACTION     ("fls",                   fls),	     /* GNU */
   PARSE_POSOPT     ("follow",                follow),  /* GNU, Unix */
@@ -265,9 +268,13 @@ static struct parser_table const parse_table[] =
   {ARG_ACTION,      "printf",   parse_printf, NULL},	     /* GNU */
   PARSE_ACTION     ("prune",                 prune),
   PARSE_ACTION     ("quit",                  quit),	     /* GNU */
+  PARSE_TEST       ("readable",              readable),	     /* GNU, 4.3.0+ */
   PARSE_TEST       ("regex",                 regex),	     /* GNU */
   PARSE_OPTION     ("regextype",             regextype),     /* GNU */
   PARSE_TEST       ("samefile",              samefile),	     /* GNU */
+#if 0
+  PARSE_OPTION     ("show-control-chars",    show_control_chars), /* GNU, 4.3.0+ */
+#endif
   PARSE_TEST       ("size",                  size),
   PARSE_TEST       ("type",                  type),
   PARSE_TEST       ("uid",                   uid),	     /* GNU */
@@ -275,6 +282,7 @@ static struct parser_table const parse_table[] =
   PARSE_TEST       ("user",                  user),
   PARSE_OPTION     ("warn",                  warn),	     /* GNU */
   PARSE_TEST_NP    ("wholename",             wholename), /* GNU, replaces -path */
+  PARSE_TEST       ("writable",              writable),	     /* GNU, 4.3.0+ */
   PARSE_OPTION     ("xdev",                  xdev),
   PARSE_TEST       ("xtype",                 xtype),	     /* GNU */
 #ifdef UNIMPLEMENTED_UNIX
@@ -303,6 +311,40 @@ static const char *first_nonoption_arg = NULL;
 
 
 
+
+void 
+set_follow_state(enum SymlinkOption opt)
+{
+  switch (opt)
+    {
+    case SYMLINK_ALWAYS_DEREF:  /* -L */
+      options.xstat = optionl_stat;
+      options.no_leaf_check = true;
+      break;
+      
+    case SYMLINK_NEVER_DEREF:	/* -P (default) */
+      options.xstat = optionp_stat;
+      /* Can't turn no_leaf_check off because the user might have specified 
+       * -noleaf anyway
+       */
+      break;
+      
+    case SYMLINK_DEREF_ARGSONLY: /* -H */
+      options.xstat = optionh_stat;
+      options.no_leaf_check = true;
+    }
+
+  options.symlink_handling = opt;
+  
+  /* For DEBUG_STAT, the choice is made at runtime within debug_stat()
+   * by checking the contents of the symlink_handling variable.
+   */
+#if defined(DEBUG_STAT)
+  options.xstat = debug_stat;
+#endif /* !DEBUG_STAT */
+}
+
+
 void
 parse_begin_user_args (char **args, int argno, const struct predicate *last, const struct predicate *predicates)
 {
@@ -471,7 +513,7 @@ parse_atime (const struct parser_table* entry, char **argv, int *arg_ptr)
 }
 
 boolean
-parse_closeparen (const struct parser_table* entry, char **argv, int *arg_ptr)
+parse_close (const struct parser_table* entry, char **argv, int *arg_ptr)
 {
   struct predicate *our_pred;
 
@@ -479,9 +521,9 @@ parse_closeparen (const struct parser_table* entry, char **argv, int *arg_ptr)
   (void) arg_ptr;
   
   our_pred = get_new_pred (entry);
-  our_pred->pred_func = pred_closeparen;
+  our_pred->pred_func = pred_close;
 #ifdef	DEBUG
-  our_pred->p_name = find_pred_name (pred_closeparen);
+  our_pred->p_name = find_pred_name (pred_close);
 #endif	/* DEBUG */
   our_pred->p_type = CLOSE_PAREN;
   our_pred->p_prec = NO_PREC;
@@ -798,6 +840,7 @@ tests (N can be +N or -N or N): -amin N -anewer FILE -atime N -cmin N\n\
       -links N -lname PATTERN -mmin N -mtime N -name PATTERN -newer FILE"));
   puts (_("\
       -nouser -nogroup -path PATTERN -perm [+-]MODE -regex PATTERN\n\
+      -readable -writable -executable\n\
       -wholename PATTERN -size N[bcwkMG] -true -type [bcdpflsD] -uid N\n\
       -used N -user NAME -xtype [bcdpfls]\n"));
   puts (_("\
@@ -847,10 +890,10 @@ fnmatch_sanitycheck(void)
 }
 
 
-static boolean
+static boolean 
 check_name_arg(const char *pred, const char *arg)
 {
-  if (options.warnings && strchr(arg, '/'))
+  if (strchr(arg, '/'))
     {
       error(0, 0,_("warning: Unix filenames usually don't contain slashes (though pathnames do).  That means that '%s %s' will probably evaluate to false all the time on this system.  You might find the '-wholename' test more useful, or perhaps '-samefile'.  Alternatively, if you are using GNU grep, you could use 'find ... -print0 | grep -FzZ %s'."),
 	    pred, arg, arg);
@@ -1205,7 +1248,7 @@ parse_okdir (const struct parser_table* entry, char **argv, int *arg_ptr)
 }
 
 boolean
-parse_openparen (const struct parser_table* entry, char **argv, int *arg_ptr)
+parse_open (const struct parser_table* entry, char **argv, int *arg_ptr)
 {
   struct predicate *our_pred;
 
@@ -1213,9 +1256,9 @@ parse_openparen (const struct parser_table* entry, char **argv, int *arg_ptr)
   (void) arg_ptr;
   
   our_pred = get_new_pred_chk_op (entry);
-  our_pred->pred_func = pred_openparen;
+  our_pred->pred_func = pred_open;
 #ifdef	DEBUG
-  our_pred->p_name = find_pred_name (pred_openparen);
+  our_pred->p_name = find_pred_name (pred_open);
 #endif	/* DEBUG */
   our_pred->p_type = OPEN_PAREN;
   our_pred->p_prec = NO_PREC;
@@ -1273,7 +1316,7 @@ parse_wholename (const struct parser_table* entry, char **argv, int *arg_ptr)
 static boolean
 parse_perm (const struct parser_table* entry, char **argv, int *arg_ptr)
 {
-  mode_t perm_val[2];
+  mode_t perm_val;
   int mode_start = 0;
   boolean havekind = false;
   enum permissions_type kind = PERM_EXACT;
@@ -1336,8 +1379,7 @@ parse_perm (const struct parser_table* entry, char **argv, int *arg_ptr)
       if (NULL == change)
 	error (1, 0, _("invalid mode `%s'"), argv[*arg_ptr]);
     }
-  perm_val[0] = mode_adjust (0, false, 0, change, NULL);
-  perm_val[1] = mode_adjust (0, true, 0, change, NULL);
+  perm_val = mode_adjust (0, change, 0);
   free (change);
   
   our_pred = insert_primary (entry);
@@ -1362,8 +1404,7 @@ parse_perm (const struct parser_table* entry, char **argv, int *arg_ptr)
 	  break;
 	}
     }
-  
-  if (('/' == argv[*arg_ptr][0]) && (0 == perm_val[0]) && (0 == perm_val[1]))
+  if (('/' == argv[*arg_ptr][0]) && (0 == perm_val))
     {
       /* The meaning of -perm /000 will change in the future.
        * It currently matches no files, but like -perm -000 it
@@ -1378,7 +1419,7 @@ parse_perm (const struct parser_table* entry, char **argv, int *arg_ptr)
 	     argv[*arg_ptr]);
     }
 
-  memcpy (our_pred->args.perm.val, perm_val, sizeof perm_val);
+  our_pred->args.perm.val = perm_val & MODE_ALL;
   (*arg_ptr)++;
   return true;
 }
@@ -1602,6 +1643,42 @@ parse_samefile (const struct parser_table* entry, char **argv, int *arg_ptr)
   return true;
 }
 
+#if 0
+static boolean
+parse_show_control_chars (const struct parser_table* entry, char **argv, int *arg_ptr)
+{
+  const char *arg;
+  const char *errmsg = _("The -show-control-chars option takes a single argument which "
+			 "must be 'literal' or 'safe'");
+  
+  if ((argv == NULL) || (argv[*arg_ptr] == NULL))
+    {
+      error (1, errno, "%s", errmsg);
+      return false;
+    }
+  else 
+    {
+      arg = argv[*arg_ptr];
+      
+      if (0 == strcmp("literal", arg))
+	{
+	  options.literal_control_chars = true;
+	}
+      else if (0 == strcmp("safe", arg))
+	{
+	  options.literal_control_chars = false;
+	}
+      else
+	{
+	  error (1, errno, "%s", errmsg);
+	  return false;
+	}
+      (*arg_ptr)++;		/* consume the argument. */
+      return true;
+    }
+}
+#endif
+
 
 static boolean
 parse_true (const struct parser_table* entry, char **argv, int *arg_ptr)
@@ -1616,6 +1693,36 @@ parse_true (const struct parser_table* entry, char **argv, int *arg_ptr)
   return true;
 }
 
+static boolean
+insert_accesscheck (const struct parser_table* entry, char **argv, int *arg_ptr)
+{
+  struct predicate *our_pred;
+  (void) argv;
+  (void) arg_ptr;
+  our_pred = insert_primary (entry);
+  our_pred->need_stat = our_pred->need_type = false;
+  our_pred->side_effects = our_pred->no_default_print = false;
+  return true;
+}
+
+static boolean
+parse_executable (const struct parser_table* entry, char **argv, int *arg_ptr)
+{
+  return insert_accesscheck(entry, argv, arg_ptr);
+}
+
+static boolean
+parse_readable (const struct parser_table* entry, char **argv, int *arg_ptr)
+{
+  return insert_accesscheck(entry, argv, arg_ptr);
+}
+
+static boolean
+parse_writable (const struct parser_table* entry, char **argv, int *arg_ptr)
+{
+  return insert_accesscheck(entry, argv, arg_ptr);
+}
+
 static boolean
 parse_type (const struct parser_table* entry, char **argv, int *arg_ptr)
 {
@@ -1688,7 +1795,6 @@ parse_version (const struct parser_table* entry, char **argv, int *arg_ptr)
   
   fflush (stderr);
   printf (_("GNU find version %s\n"), version_string);
-  printf (_("Built using GNU gnulib version %s\n"), gnulib_version);
   printf (_("Features enabled: "));
   
 #if CACHE_IDS
@@ -1716,6 +1822,13 @@ parse_version (const struct parser_table* entry, char **argv, int *arg_ptr)
   printf("LEAF_OPTIMISATION ");
   ++features;
 #endif
+
+  if (is_fts_enabled())
+    {
+      printf("FTS ");
+      ++features;
+    }
+
   if (0 == features)
     {
       /* For the moment, leave this as English in case someone wants
@@ -1948,12 +2061,7 @@ insert_fprintf (FILE *fp, const struct parser_table *entry, PRED_FUNC func, char
 	}
       else if (*scan == '%')
 	{
-	  if (scan[1] == 0)
-	    {
-	      /* Trailing %.  We don't like those. */
-	      error (1, 0, _("error: %s at end of format string"), scan);
-	    }
-	  else if (scan[1] == '%')
+	  if (scan[1] == '%')
 	    {
 	      segmentp = make_segment (segmentp, format, scan - format + 1,
 				       KIND_PLAIN);
@@ -2115,7 +2223,6 @@ new_insert_exec_ok (const char *action,
   boolean allow_plus;		/* True if + is a valid terminator */
   int brace_count;		/* Number of instances of {}. */
   PRED_FUNC func = entry->pred_func;
-  enum BC_INIT_STATUS bcstatus;
   
   struct predicate *our_pred;
   struct exec_val *execp;	/* Pointer for efficiency. */
@@ -2218,24 +2325,8 @@ new_insert_exec_ok (const char *action,
 	    suffix);
     }
 
-  /* We use a switch statement here so that 
-   * the compiler warns us when we forget to handle a 
-   * newly invented enum value.
-   */
-  bcstatus = bc_init_controlinfo(&execp->ctl);
-  switch (bcstatus) 
-    {
-    case BC_INIT_ENV_TOO_BIG:
-      error(1, 0, 
-	    _("The environment is too large for exec()."));
-      break;
-    case BC_INIT_OK:
-      /* Good news.  Carry on. */
-      break;
-    }
-  bc_use_sensible_arg_max(&execp->ctl);
-
-
+  /* execp->ctl   = xmalloc(sizeof struct buildcmd_control); */
+  bc_init_controlinfo(&execp->ctl);
   execp->ctl.exec_callback = launch;
 
   if (our_pred->args.exec_vec.multiple)
