@@ -1,19 +1,21 @@
 /* pred.c -- execute the expression tree.
    Copyright (C) 1990, 1991, 1992, 1993, 1994, 2000, 2003, 
-                 2004, 2005, 2007 Free Software Foundation, Inc.
+                 2004, 2005 Free Software Foundation, Inc.
 
-   This program is free software: you can redistribute it and/or modify
+   This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation, either version 3 of the License, or
-   (at your option) any later version.
-   
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+   USA.
 */
 
 #include "defs.h"
@@ -35,6 +37,7 @@
 #include "printquoted.h"
 #include "buildcmd.h"
 #include "yesno.h"
+#include "listfile.h"
 
 #if ENABLE_NLS
 # include <libintl.h>
@@ -153,11 +156,9 @@ static char *format_date PARAMS((time_t when, int kind));
 static char *ctime_format PARAMS((time_t when));
 
 #ifdef	DEBUG
-typedef boolean (*PFB) (char *, struct stat *, struct predicate *);
-
 struct pred_assoc
 {
-  PFB pred_func;
+  PRED_FUNC pred_func;
   char *pred_name;
 };
 
@@ -167,7 +168,7 @@ struct pred_assoc pred_table[] =
   {pred_and, "and     "},
   {pred_anewer, "anewer  "},
   {pred_atime, "atime   "},
-  {pred_closeparen, ")       "},
+  {pred_close, ")       "},
   {pred_amin, "cmin    "},
   {pred_cnewer, "cnewer  "},
   {pred_comma, ",       "},
@@ -199,7 +200,7 @@ struct pred_assoc pred_table[] =
   {pred_nouser, "nouser  "},
   {pred_ok, "ok      "},
   {pred_okdir, "okdir   "},
-  {pred_openparen, "(       "},
+  {pred_open, "(       "},
   {pred_or, "or      "},
   {pred_path, "path    "},
   {pred_perm, "perm    "},
@@ -334,7 +335,7 @@ pred_atime (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_closeparen (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_close (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) &pathname;
   (void) &stat_buf;
@@ -527,8 +528,9 @@ boolean
 pred_fls (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   list_file (pathname, state.rel_pathname, stat_buf, options.start_time,
-	     options.output_block_size, pred_ptr->args.stream);
-  return (true);
+	     options.output_block_size,
+	     pred_ptr->literal_control_chars, pred_ptr->args.stream);
+  return true;
 }
 
 boolean
@@ -654,11 +656,7 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	  break;
 	case 'f':		/* base name of path */
 	  /* sanitised */
-	  {
-	    char *base = base_name (pathname);
-	    print_quoted (fp, qopts, ttyflag, segment->text, base);
-	    free (base);
-	  }
+	  print_quoted (fp, qopts, ttyflag, segment->text, base_name (pathname));
 	  break;
 	case 'F':		/* filesystem type */
 	  /* trusted */
@@ -713,11 +711,11 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	case 'H':		/* ARGV element file was found under */
 	  /* trusted */
 	  {
-	    char cc = pathname[state.path_length];
+	    char cc = pathname[state.starting_path_length];
 
-	    pathname[state.path_length] = '\0';
+	    pathname[state.starting_path_length] = '\0';
 	    fprintf (fp, segment->text, pathname);
-	    pathname[state.path_length] = cc;
+	    pathname[state.starting_path_length] = cc;
 	    break;
 	  }
 	case 'i':		/* inode number */
@@ -761,7 +759,7 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	  /* UNTRUSTED, probably unexploitable */
 	  {
 	    char modestring[16] ;
-	    filemodestring (stat_buf, modestring);
+	    mode_string (stat_buf->st_mode, modestring);
 	    modestring[10] = '\0';
 	    fprintf (fp, segment->text, modestring);
 	  }
@@ -814,7 +812,7 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	  /* sanitised */
 	  if (state.curdepth > 0)
 	    {
-	      cp = pathname + state.path_length;
+	      cp = pathname + state.starting_path_length;
 	      if (*cp == '/')
 		/* Move past the slash between the ARGV element
 		   and the rest of the pathname.  But if the ARGV element
@@ -963,39 +961,20 @@ pred_ilname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
   return insert_lname (pathname, stat_buf, pred_ptr, true);
 }
 
-/* Common code between -name, -iname.  PATHNAME is being visited, STR
-   is name to compare basename against, and FLAGS are passed to
-   fnmatch.  Recall that 'find / -name /' is one of the few times where a '/' 
-   in the -name must actually find something. */ 
-static boolean
-pred_name_common (const char *pathname, const char *str, int flags)
+boolean
+pred_iname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  char *p;
-  boolean b;
-  /* We used to use last_component() here, but that would not allow us
-   * to modify the input string, which is const.  We could optimise by
-   * duplicating the string only if we need to modify it, and I'll do
-   * that if there is a measurable performance difference on a machine
-   * built after 1990...
-   */
-  char *base = base_name (pathname);
-  /* remove trailing slashes, but leave  "/" or "//foo" unchanged. */
-  strip_trailing_slashes(base);
+  const char *base;
+
+  (void) stat_buf;
 
   /* FNM_PERIOD is not used here because POSIX requires that it not be.
    * See http://standards.ieee.org/reading/ieee/interp/1003-2-92_int/pasc-1003.2-126.html
    */
-  b = fnmatch (str, base, flags) == 0;
-  free (base);
-  return b;
-}
-
-
-boolean
-pred_iname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
-{
-  (void) stat_buf;
-  return pred_name_common (pathname, pred_ptr->args.str, FNM_CASEFOLD);
+  base = base_name (pathname);
+  if (fnmatch (pred_ptr->args.str, base, FNM_CASEFOLD) == 0)
+    return (true);
+  return (false);
 }
 
 boolean
@@ -1083,11 +1062,11 @@ insert_lname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr,
 boolean
 pred_ls (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  (void) pred_ptr;
-  
   list_file (pathname, state.rel_pathname, stat_buf, options.start_time,
-	     options.output_block_size, stdout);
-  return (true);
+	     options.output_block_size,
+	     pred_ptr->literal_control_chars,
+	     stdout);
+  return true;
 }
 
 boolean
@@ -1107,8 +1086,17 @@ pred_mtime (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_name (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  const char *base;
+
   (void) stat_buf;
-  return pred_name_common (pathname, pred_ptr->args.str, 0);
+  base = base_name (pathname);
+
+  /* FNM_PERIOD is not used here because POSIX requires that it not be.
+   * See http://standards.ieee.org/reading/ieee/interp/1003-2-92_int/pasc-1003.2-126.html
+   */
+  if (fnmatch (pred_ptr->args.str, base, 0) == 0)
+    return (true);
+  return (false);
 }
 
 boolean
@@ -1201,7 +1189,7 @@ pred_okdir (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 }
 
 boolean
-pred_openparen (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+pred_open (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pathname;
   (void) stat_buf;
@@ -1237,13 +1225,11 @@ pred_path (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_perm (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  mode_t mode = stat_buf->st_mode;
-  mode_t perm_val = pred_ptr->args.perm.val[S_ISDIR (mode) != 0];
   (void) pathname;
   switch (pred_ptr->args.perm.kind)
     {
     case PERM_AT_LEAST:
-      return (mode & perm_val) == perm_val;
+      return (stat_buf->st_mode & pred_ptr->args.perm.val) == pred_ptr->args.perm.val;
       break;
 
     case PERM_ANY:
@@ -1257,14 +1243,14 @@ pred_perm (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
        * Eric Blake's interpretation is that the mode argument is zero, 
        
        */
-      if (0 == perm_val)
+      if (0 == pred_ptr->args.perm.val)
 	return true;		/* Savannah bug 14748; we used to return false */
       else
-	return (mode & perm_val) != 0;
+	return (stat_buf->st_mode & pred_ptr->args.perm.val) != 0;
       break;
 
     case PERM_EXACT:
-      return (mode & MODE_ALL) == perm_val;
+      return (stat_buf->st_mode & MODE_ALL) == pred_ptr->args.perm.val;
       break;
 
     default:
@@ -1273,6 +1259,25 @@ pred_perm (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
     }
 }
 
+
+boolean
+pred_executable (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+  return 0 == access(state.rel_pathname, X_OK);
+}
+
+boolean
+pred_readable (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+  return 0 == access(state.rel_pathname, R_OK);
+}
+
+boolean
+pred_writable (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+  return 0 == access(state.rel_pathname, W_OK);
+}
+
 boolean
 pred_print (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
@@ -1715,15 +1720,14 @@ ctime_format (time_t when)
    the predicate function PRED_FUNC. */
 
 char *
-find_pred_name (pred_func)
-     PFB pred_func;
+find_pred_name (PRED_FUNC func)
 {
   int i;
 
   for (i = 0; pred_table[i].pred_func != 0; i++)
-    if (pred_table[i].pred_func == pred_func)
+    if (pred_table[i].pred_func == func)
       break;
-  return (pred_table[i].pred_name);
+  return pred_table[i].pred_name;
 }
 
 static char *
@@ -1949,3 +1953,5 @@ pred_sanity_check(const struct predicate *predicates)
 	}
     }
 }
+
+
