@@ -1,26 +1,45 @@
 /* listfile.c -- display a long listing of a file
-   Copyright (C) 1991, 1993, 2000, 2003, 2004, 2007 Free Software
-   Foundation, Inc.
+   Copyright (C) 1991, 1993, 2000 Free Software Foundation, Inc.
 
-   This program is free software: you can redistribute it and/or modify
+   This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation, either version 3 of the License, or
-   (at your option) any later version.
-   
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+   USA.
 */
 
 #if HAVE_CONFIG_H
 # include <config.h>
 #endif
 
-#include <alloca.h>
+#ifndef __GNUC__
+# if HAVE_ALLOCA_H
+#  include <alloca.h>
+# else
+#  ifdef _AIX
+ #  pragma alloca
+#  else
+#   ifdef _WIN32
+#    include <malloc.h>
+#    include <io.h>
+#   else
+#    ifndef alloca
+char *alloca ();
+#    endif
+#   endif
+#  endif
+# endif
+#endif
+
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <stdio.h>
@@ -33,7 +52,6 @@
 #include "pathmax.h"
 #include "error.h"
 #include "filemode.h"
-#include "idcache.h"
 
 #include "listfile.h"
 
@@ -44,8 +62,8 @@
 #endif
 
 
-/* The presence of unistd.h is assumed by gnulib these days, so we
- * might as well assume it too.
+/* The presence of unistd.h is assumed by gnulib these days, so we 
+ * might as well assume it too. 
  */
 #include <unistd.h> /* for readlink() */
 
@@ -160,7 +178,10 @@ struct group *getgrgid ();
 
 
 char * get_link_name (char *name, char *relname);
-static void print_name_with_quoting (register char *p, FILE *stream);
+static void print_name (register char *p, FILE *stream, int literal_control_chars);
+
+extern char * getgroup (gid_t gid);
+extern char * getuser (uid_t uid);
 
 
 /* NAME is the name to print.
@@ -177,9 +198,10 @@ list_file (char *name,
 	   struct stat *statp,
 	   time_t current_time,
 	   int output_block_size,
+	   int literal_control_chars,
 	   FILE *stream)
 {
-  char modebuf[12];
+  char modebuf[11];
   struct tm const *when_local;
   char const *user_name;
   char const *group_name;
@@ -187,10 +209,11 @@ list_file (char *name,
 
 #if HAVE_ST_DM_MODE
   /* Cray DMF: look at the file's migrated, not real, status */
-  strmode (statp->st_dm_mode, modebuf);
+  mode_string (statp->st_dm_mode, modebuf);
 #else
-  strmode (statp->st_mode, modebuf);
+  mode_string (statp->st_mode, modebuf);
 #endif
+  modebuf[10] = '\0';
 
   fprintf (stream, "%6s ",
 	   human_readable ((uintmax_t) statp->st_ino, hbuf,
@@ -203,9 +226,9 @@ list_file (char *name,
 			   ST_NBLOCKSIZE, output_block_size));
 
 
-  /* modebuf includes the space between the mode and the number of links,
-     as the POSIX "optional alternate access method flag".  */
-  fprintf (stream, "%s%3lu ", modebuf, (unsigned long) statp->st_nlink);
+  /* The space between the mode and the number of links is the POSIX
+     "optional alternate access method flag".  */
+  fprintf (stream, "%s %3lu ", modebuf, (unsigned long) statp->st_nlink);
 
   user_name = getuser (statp->st_uid);
   if (user_name)
@@ -229,7 +252,7 @@ list_file (char *name,
 #endif
   else
     fprintf (stream, "%8s ",
-	     human_readable ((uintmax_t) statp->st_size, hbuf,
+	     human_readable ((uintmax_t) statp->st_size, hbuf, 
 			     human_ceiling,
 			     1,
 			     output_block_size < 0 ? output_block_size : 1));
@@ -280,7 +303,7 @@ list_file (char *name,
 				 1, 1));
     }
 
-  print_name_with_quoting (name, stream);
+  print_name (name, stream, literal_control_chars);
 
 #ifdef S_ISLNK
   if (S_ISLNK (statp->st_mode))
@@ -290,7 +313,7 @@ list_file (char *name,
       if (linkname)
 	{
 	  fputs (" -> ", stream);
-	  print_name_with_quoting (linkname, stream);
+	  print_name (linkname, stream, literal_control_chars);
 	  free (linkname);
 	}
     }
@@ -298,6 +321,14 @@ list_file (char *name,
   putc ('\n', stream);
 }
 
+
+static void
+print_name_without_quoting (char *p, FILE *stream)
+{
+  fprintf(stream, "%s", p);
+}
+
+
 static void
 print_name_with_quoting (register char *p, FILE *stream)
 {
@@ -348,6 +379,14 @@ print_name_with_quoting (register char *p, FILE *stream)
     }
 }
 
+static void print_name (register char *p, FILE *stream, int literal_control_chars)
+{
+  if (literal_control_chars)
+    print_name_without_quoting(p, stream);
+  else
+    print_name_with_quoting(p, stream);
+}
+
 #ifdef S_ISLNK
 char *
 get_link_name (char *name, char *relname)
