@@ -1,23 +1,27 @@
 #! /bin/sh
 # updatedb -- build a locate pathname database
-# Copyright (C) 1994, 1996, 1997, 2000, 2001, 2003, 2004, 2005, 2006
-# Free Software Foundation, Inc.
-#
-# This program is free software: you can redistribute it and/or modify
+# Copyright (C) 1994 Free Software Foundation, Inc.
+
+# This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-# 
+# the Free Software Foundation; either version 2, or (at your option)
+# any later version.
+
 # This program is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
-# 
+
 # You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+# USA.
 
 # csh original by James Woods; sh conversion by David MacKenzie.
 
+#exec 2> /tmp/updatedb-trace.txt 
+#set -x
+
 usage="\
 Usage: $0 [--findoptions='-option1 -option2...']
        [--localpaths='dir1 dir2...'] [--netpaths='dir1 dir2...']
@@ -83,11 +87,11 @@ getuid() {
 
 # figure out if su supports the -s option
 select_shell() {
-    if su "$1" -s $SHELL -c false < /dev/null  ; then
+    if su "$1" -s $SHELL false < /dev/null  ; then
 	# No.
 	echo ""
     else
-	if su "$1" -s $SHELL -c true < /dev/null  ; then
+	if su "$1" -s $SHELL true < /dev/null  ; then
 	    # Yes.
 	    echo "-s $SHELL"
         else
@@ -168,6 +172,21 @@ fi
 : ${code:=${LIBEXECDIR}/@code@}
 
 
+checkbinary () {
+    if test -x "$1" ; then
+	: ok
+    else
+      eval echo "updatedb needs to be able to execute $1, but cannot." >&2
+      exit 1
+    fi
+}
+
+for binary in $find $frcode $bigram $code
+do
+  checkbinary $binary
+done
+
+
 PATH=/bin:/usr/bin:${BINDIR}; export PATH
 
 : ${PRUNEFS="nfs NFS proc afs proc smbfs autofs iso9660 ncpfs coda devpts ftpfs devfs mfs sysfs shfs"}
@@ -220,7 +239,7 @@ if [ "$myuid" = 0 ]; then
 fi
 } | $sort -f | $frcode $frcode_options > $LOCATE_DB.n
 then
-    # OK so far
+    : OK so far
     true
 else
     rv=$?
