@@ -1,24 +1,26 @@
 /* xargs -- build and execute command lines from standard input
-   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003, 2004, 2005, 2006, 2007 Free Software Foundation, Inc.
+   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003, 2005 Free Software Foundation, Inc.
 
-   This program is free software: you can redistribute it and/or modify
+   This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation, either version 3 of the License, or
-   (at your option) any later version.
-   
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+   USA.
 */
 
 /* Written by Mike Rendell <michael@cs.mun.ca>
-   and David MacKenzie <djm@gnu.org>.
-   Modifications by
-	James Youngman
+   and David MacKenzie <djm@gnu.org>.  
+   Modifications by 
+   	James Youngman
 	Dmitry V. Levin
 */
 
@@ -131,6 +133,8 @@ extern int errno;
 /* Return nonzero if S is the EOF string.  */
 #define EOF_STR(s) (eof_str && *eof_str == *s && !strcmp (eof_str, s))
 
+extern char **environ;
+
 /* Do multibyte processing if multibyte characters are supported,
    unless multibyte sequences are search safe.  Multibyte sequences
    are search safe if searching for a substring using the byte
@@ -165,8 +169,7 @@ typedef int boolean;
 #endif
 
 #include <xalloc.h>
-#include "closein.h"
-#include "gnulib-version.h"
+#include "closeout.h"
 
 void error PARAMS ((int status, int errnum, char *message,...));
 
@@ -220,9 +223,9 @@ static int pids_alloc = 0;
 
 /* Exit status; nonzero if any child process exited with a
    status of 1-125.  */
-static volatile int child_error = 0;
+volatile static int child_error = 0;
 
-static volatile int original_exit_value;
+volatile static int original_exit_value;
 
 /* If true, print each command on stderr before executing it.  */
 static boolean print_command = false; /* Option -t */
@@ -268,10 +271,37 @@ static void add_proc PARAMS ((pid_t pid));
 static void wait_for_proc PARAMS ((boolean all));
 static void wait_for_proc_all PARAMS ((void));
 static long parse_num PARAMS ((char *str, int option, long min, long max, int fatal));
+static long env_size PARAMS ((char **envp));
 static void usage PARAMS ((FILE * stream));
 
 
 
+static long
+get_line_max(void)
+{
+  long val;
+#ifdef _SC_LINE_MAX  
+  val = sysconf(_SC_LINE_MAX);
+#else
+  val = -1;
+#endif
+  
+  if (val > 0)
+    return val;
+
+  /* either _SC_LINE_MAX was not available or 
+   * there is no particular limit.
+   */
+#ifdef LINE_MAX
+  val = LINE_MAX;
+#endif
+
+  if (val > 0)
+    return val;
+
+  return 2048L;			/* a reasonable guess. */
+}
+
 static char 
 get_char_oct_or_hex_escape(const char *s)
 {
@@ -341,6 +371,8 @@ get_char_oct_or_hex_escape(const char *s)
 static char 
 get_input_delimiter(const char *s)
 {
+  char result = '\0';
+  
   if (1 == strlen(s))
     {
       return s[0];
@@ -377,22 +409,10 @@ get_input_delimiter(const char *s)
 	  error(1, 0,
 		_("Invalid input delimiter specification %s: the delimiter must be either a single character or an escape sequence starting with \\."),
 		s);
-	  abort ();
 	}
     }
 }
 
-static void 
-noop (void)
-{
-  /* does nothing. */
-}
-
-static void 
-fail_due_to_env_size (void)
-{
-  error (1, 0, _("environment is too large for exec"));
-}
 
 
 int
@@ -402,10 +422,13 @@ main (int argc, char **argv)
   int show_limits = 0;			/* --show-limits */
   int always_run_command = 1;
   char *input_file = "-"; /* "-" is stdin */
+  long posix_arg_size_max;
+  long posix_arg_size_min;
+  long arg_size;
+  long size_of_environment = env_size(environ);
   char *default_cmd = "/bin/echo";
   int (*read_args) PARAMS ((void)) = read_line;
-  void (*act_on_init_result)(void) = noop;
-  enum BC_INIT_STATUS bcstatus;
+  int env_too_big = 0;
 
   program_name = argv[0];
   original_exit_value = 0;
@@ -415,44 +438,47 @@ main (int argc, char **argv)
 #endif
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
-  atexit (close_stdin);
+  atexit (close_stdout);
   atexit (wait_for_proc_all);
 
-  bcstatus = bc_init_controlinfo(&bc_ctl);
-  /* The bc_init_controlinfo call may have determined that the 
-   * environment is too big.  In that case, we will fail with 
-   * an error message after processing the command-line options,
-   * as "xargs --help" should still work even if the environment is 
-   * too big.
-   *
-   * Some of the argument processing depends on the contents of 
-   * bc_ctl, which will be in an undefined state if bc_init_controlinfo()
-   * failed.
+  /* IEEE Std 1003.1, 2003 specifies that the combined argument and 
+   * environment list shall not exceed {ARG_MAX}-2048 bytes.  It also 
+   * specifies that it shall be at least LINE_MAX.
    */
-  if (BC_INIT_ENV_TOO_BIG == bcstatus)
+  posix_arg_size_min = get_line_max();
+  posix_arg_size_max = bc_get_arg_max();
+  posix_arg_size_max -= 2048; /* POSIX.2 requires subtracting 2048.  */
+
+  bc_init_controlinfo(&bc_ctl);
+  assert(bc_ctl.arg_max == posix_arg_size_max);
+
+  bc_ctl.exec_callback = xargs_do_exec;
+
+  
+  /* Start with a reasonable default size, though this can be
+   * adjusted via the -s option.
+   */
+  arg_size = (128 * 1024) + size_of_environment;
+
+  /* Take the size of the environment into account.  */
+  if (size_of_environment > posix_arg_size_max)
     {
-      act_on_init_result = fail_due_to_env_size;
+      bc_ctl.arg_max = 0;
+      env_too_big = 1;
     }
   else
     {
-      /* IEEE Std 1003.1, 2003 specifies that the combined argument and 
-       * environment list shall not exceed {ARG_MAX}-2048 bytes.  It also 
-       * specifies that it shall be at least LINE_MAX.
-       */
-#if defined(ARG_MAX)
-      assert(bc_ctl.arg_max <= (ARG_MAX-2048));
-#endif
-#ifdef LINE_MAX
-      assert(bc_ctl.arg_max >= LINE_MAX);
-#endif
-      
-      bc_ctl.exec_callback = xargs_do_exec;
-      
-      /* Start with a reasonable default size, though this can be
-       * adjusted via the -s option.
-       */
-      bc_use_sensible_arg_max(&bc_ctl);
+      bc_ctl.arg_max = posix_arg_size_max - size_of_environment;
     }
+
+  /* Check against the upper and lower limits. */  
+  if (arg_size > bc_ctl.arg_max)
+    arg_size = bc_ctl.arg_max;
+  if (arg_size < posix_arg_size_min)
+    arg_size = posix_arg_size_min;
+  
+  
+
   
   while ((optc = getopt_long (argc, argv, "+0a:E:e::i::I:l::L:n:prs:txP:d:",
 			      longopts, (int *) 0)) != -1)
@@ -525,21 +551,12 @@ main (int argc, char **argv)
 	   * cannot support - in that case, the relevant limit is used.
 	   */
 	case 's':
-	  {
-	    size_t arg_size;
-	    act_on_init_result();
-	    arg_size = parse_num (optarg, 's', 1L,
-				  bc_ctl.posix_arg_size_max, 0);
-	    if (arg_size > bc_ctl.posix_arg_size_max)
-	      {
-		error (0, 0,
-		       _("warning: value %ld for -s option is too large, "
-			 "using %ld instead"),
-		       arg_size, bc_ctl.posix_arg_size_max);
-		arg_size = bc_ctl.posix_arg_size_max;
-	      }
-	    bc_ctl.arg_max = arg_size;
-	  }
+	  arg_size = parse_num (optarg, 's', 1L, posix_arg_size_max, 0);
+	  if (arg_size > posix_arg_size_max)
+	    {
+	      error (0, 0, "warning: value %ld for -s option is too large, using %ld instead", arg_size, posix_arg_size_max);
+	      arg_size = posix_arg_size_max;
+	    }
 	  break;
 
 	case 'S':
@@ -573,7 +590,6 @@ main (int argc, char **argv)
 
 	case 'v':
 	  printf (_("GNU xargs version %s\n"), version_string);
-	  printf (_("Built using GNU gnulib version %s\n"), gnulib_version);
 	  return 0;
 
 	default:
@@ -582,15 +598,14 @@ main (int argc, char **argv)
 	}
     }
 
-  /* If we had deferred failing due to problems in bc_init_controlinfo(),
-   * do it now.
-   *
-   * We issue this error message after processing command line 
-   * arguments so that it is possible to use "xargs --help" even if
-   * the environment is too large. 
-   */
-  act_on_init_result();
-  assert(BC_INIT_OK == bcstatus);
+  if (env_too_big)
+  {
+    /* We issue this error message after processing command line 
+     * arguments so that it is possible to use "xargs --help" even if
+     * the environment is too large. 
+     */
+    error (1, 0, _("environment is too large for exec"));
+  }
 
   if (0 == strcmp (input_file, "-"))
     {
@@ -618,41 +633,39 @@ main (int argc, char **argv)
       argv = &default_cmd;
     }
 
-  /* We want to be able to print size_t values as unsigned long, so if
-   * the cast isn't value-preserving, we have a problem.  This isn't a
-   * problem in C89, because size_t was known to be no wider than
-   * unsigned long.  In C99 this is no longer the case, but there are
-   * special C99 ways to print such values.  Unfortunately this
-   * program tries to work on both C89 and C99 systems.
+  /* Taking into account the size of the environment, 
+   * figure out how large a buffer we need to
+   * hold all the arguments.  We cannot use ARG_MAX 
+   * directly since that may be arbitrarily large.
+   * This is from a patch by Bob Proulx, <bob@proulx.com>.
    */
-#if defined(SIZE_MAX)
-# if SIZE_MAX > ULONG_MAX
-# error "I'm not sure how to print size_t values on your system"
-# endif
-#else
-  /* Without SIZE_MAX (i.e. limits.h) this is probably 
-   * close to the best we can do.
-   */
-  assert(sizeof(size_t) <= sizeof(unsigned long));
-#endif
+  if (bc_ctl.arg_max > arg_size)
+    {
+      if (show_limits)
+	{
+	  fprintf(stderr,
+		  _("Reducing arg_max (%ld) to arg_size (%ld)\n"),
+		  bc_ctl.arg_max, arg_size);
+	}
+      bc_ctl.arg_max = arg_size;
+    }
 
   if (show_limits)
     {
       fprintf(stderr,
-	      _("Your environment variables take up %lu bytes\n"),
-	      (unsigned long)bc_size_of_environment());
+	      _("Your environment variables take up %ld bytes\n"),
+	      size_of_environment);
       fprintf(stderr,
-	      _("POSIX lower and upper limits on argument length: %lu, %lu\n"),
-	      (unsigned long)bc_ctl.posix_arg_size_min,
-	      (unsigned long)bc_ctl.posix_arg_size_max);
+	      _("POSIX lower and upper limits on argument length: %ld, %ld\n"),
+	      posix_arg_size_min,
+	      posix_arg_size_max);
       fprintf(stderr,
 	      _("Maximum length of command we could actually use: %ld\n"),
-	      (unsigned long)(bc_ctl.posix_arg_size_max -
-			      bc_size_of_environment()));
+	      (posix_arg_size_max - size_of_environment));
       fprintf(stderr,
-	      _("Size of command buffer we are actually using: %lu\n"),
-	      (unsigned long)bc_ctl.arg_max);
-
+	      _("Size of command buffer we are actually using: %ld\n"),
+	      bc_ctl.arg_max);
+      
       if (isatty(STDIN_FILENO))
 	{
 	  fprintf(stderr,
@@ -662,6 +675,7 @@ main (int argc, char **argv)
 		  "not what you wanted to happen, please type the "
 		  "end-of-file keystroke.\n");
 	}
+      
     }
   
   linebuf = (char *) xmalloc (bc_ctl.arg_max + 1);
@@ -1209,6 +1223,17 @@ parse_num (char *str, int option, long int min, long int max, int fatal)
 
 /* Return how much of ARG_MAX is used by the environment.  */
 
+static long
+env_size (char **envp)
+{
+  long len = 0;
+
+  while (*envp)
+    len += strlen (*envp++) + 1;
+
+  return len;
+}
+
 static void
 usage (FILE *stream)
 {
@@ -1219,7 +1244,7 @@ Usage: %s [-0prtx] [--interactive] [--null] [-d|--delimiter=delim]\n\
        [-I replace-str] [-i[replace-str]] [--replace[=replace-str]]\n\
        [-n max-args] [--max-args=max-args]\n\
        [-s max-chars] [--max-chars=max-chars]\n\
-       [-P max-procs]  [--max-procs=max-procs] [--show-limits]\n\
+       [-P max-procs]  [--max-procs=max-procs]\n\
        [--verbose] [--exit] [--no-run-if-empty] [--arg-file=file]\n\
        [--version] [--help] [command [initial-arguments]]\n"),
 	   program_name);
